use std::ops::DerefMut;
use std::sync::Mutex;

use eframe::egui::{self};

use plate_tool_lib::plate::PlateFormat;
use plate_tool_lib::plate_instances;

use crate::plate::{add_grid, PlateUiState};
use crate::transfer_menu::{self, transfer_menu, CurrentTransferState, TransferMenuState};

#[derive(Default)]
pub struct MainState {
    pub source_plates: Vec<plate_tool_lib::plate_instances::PlateInstance>,
    pub destination_plates: Vec<plate_tool_lib::plate_instances::PlateInstance>,
    pub transfers: Vec<plate_tool_lib::transfer::Transfer>,
}

fn construct_fake_mainstate() -> MainState {
    let src_plate: plate_tool_lib::plate_instances::PlateInstance =
        plate_instances::PlateInstance::new(
            plate_tool_lib::plate::PlateType::Source,
            PlateFormat::W96,
            "src1".to_owned(),
        );
    let dest_plate: plate_tool_lib::plate_instances::PlateInstance =
        plate_instances::PlateInstance::new(
            plate_tool_lib::plate::PlateType::Destination,
            PlateFormat::W96,
            "dest1".to_owned(),
        );
    let well_a1 = plate_tool_lib::Well { row: 1, col: 1 };
    let well_c3 = plate_tool_lib::Well { row: 3, col: 3 };
    let well_a5 = plate_tool_lib::Well { row: 1, col: 5 };

    let transfer1_region: plate_tool_lib::transfer_region::TransferRegion =
        plate_tool_lib::transfer_region::TransferRegion {
            source_plate: src_plate.plate,
            source_region: plate_tool_lib::transfer_region::Region::Rect(well_a1, well_c3),
            dest_plate: dest_plate.plate,
            dest_region: plate_tool_lib::transfer_region::Region::Point(well_a1),
            interleave_source: (1, 1),
            interleave_dest: (1, 1),
        };
    let transfer1: plate_tool_lib::transfer::Transfer = plate_tool_lib::transfer::Transfer::new(
        src_plate.clone(),
        dest_plate.clone(),
        transfer1_region,
        "Shrimp".to_owned(),
    );

    let transfer2_region: plate_tool_lib::transfer_region::TransferRegion =
        plate_tool_lib::transfer_region::TransferRegion {
            source_plate: src_plate.plate,
            source_region: plate_tool_lib::transfer_region::Region::Rect(well_a1, well_c3),
            dest_plate: dest_plate.plate,
            dest_region: plate_tool_lib::transfer_region::Region::Point(well_a5),
            interleave_source: (1, 1),
            interleave_dest: (2, 2),
        };
    let transfer2: plate_tool_lib::transfer::Transfer = plate_tool_lib::transfer::Transfer::new(
        src_plate.clone(),
        dest_plate.clone(),
        transfer2_region,
        "Shrimp".to_owned(),
    );


    MainState {
        source_plates: vec![src_plate],
        destination_plates: vec![dest_plate],
        transfers: vec![transfer1, transfer2],
    }
}

#[non_exhaustive]
struct MainWindowState {
    show_side_panel: bool,
}

impl Default for MainWindowState {
    fn default() -> Self {
        Self {
            show_side_panel: true,
        }
    }
}

#[non_exhaustive]
pub struct PlateToolEframe {
    source_plate_state: Mutex<PlateUiState>,
    destination_plate_state: Mutex<PlateUiState>,
    main_window_state: MainWindowState,
    current_transfer_state: CurrentTransferState,
    transfer_menu_state: TransferMenuState,
    transfer_region_cache: plate_tool_lib::transfer_region_cache::TransferRegionCache,
    main_state: MainState,
}

impl Default for PlateToolEframe {
    fn default() -> Self {
        Self {
            source_plate_state: Mutex::new(PlateUiState::default()),
            destination_plate_state: Mutex::new(PlateUiState::default()),
            main_window_state: MainWindowState::default(),
            current_transfer_state: CurrentTransferState::default(),
            transfer_menu_state: TransferMenuState::default(),
            transfer_region_cache: plate_tool_lib::transfer_region_cache::TransferRegionCache::new(
            ),
            main_state: construct_fake_mainstate(),
        }
    }
}

impl PlateToolEframe {
    pub fn new(cc: &eframe::CreationContext<'_>) -> Self {
        // Would load state here

        let pte: PlateToolEframe = Default::default();
        pte.transfer_region_cache.generate_cache(&pte.main_state.transfers);

        pte
    }
}

impl eframe::App for PlateToolEframe {
    // State storage
    /*
    fn save(&mut self, storage: &mut dyn eframe::Storage) {
        unimplemented!()
    };
    */

    fn update(&mut self, ctx: &eframe::egui::Context, _frame: &mut eframe::Frame) {
        let ordered_ids: Vec<plate_tool_lib::uuid::Uuid> = {
            let mut ids: Vec<plate_tool_lib::uuid::Uuid> = self
                .main_state
                .transfers
                .clone()
                .iter()
                .map(|x| x.id)
                .collect();
            ids.sort_unstable();
            ids
        };

        egui::TopBottomPanel::top("top_panel").show(ctx, |ui| {
            egui::menu::bar(ui, |ui| {
                ui.menu_button("File", |ui| {
                    if ui.button("New").clicked() {}
                    ui.menu_button("Export", |ui| {
                        if ui.button("Export as CSV").clicked() {}
                        if ui.button("Export as JSON").clicked() {}
                    });
                    ui.menu_button("Import", |ui| {
                        if ui.button("Import from JSON").clicked() {}
                        if ui.button("Import transfer from CSV").clicked() {}
                    });
                });
                ui.menu_button("Options", |ui| {
                    ui.menu_button("Styles", |ui| {
                        if ui.button("Toggle transfer hashes").clicked() {}
                        if ui.button("Toggle volume heatmap").clicked() {}
                        if ui.button("Toggle current coordinates view").clicked() {}
                    });
                    ui.menu_button("Exports", |ui| {
                        if ui.button("Change CSV export type").clicked() {}
                    });
                    ui.menu_button("Windows", |ui| {
                        ui.toggle_value(
                            &mut self.main_window_state.show_side_panel,
                            "Toggle side panel",
                        );
                    });
                });
            });
        });

        if self.main_window_state.show_side_panel {
            egui::SidePanel::left("side_menus").show(ctx, |ui| {
                ui.vertical(|ui| {
                    transfer_menu(ui, &self.current_transfer_state, &mut self.transfer_menu_state);
                });
            });
        }

        egui::CentralPanel::default().show(ctx, |ui| {
            ui.vertical(|ui| {
                let available_size = ui.available_size();
                let half_height = {
                    let mut x = available_size;
                    x.y /= 2.0;
                    x
                };
                add_grid(
                    half_height,
                    PlateFormat::W96,
                    &self.main_state.transfers,
                    plate_tool_lib::plate::PlateType::Source,
                    &ordered_ids,
                    &self.transfer_region_cache,
                    Some(&self.current_transfer_state),
                    ui,
                    self.source_plate_state.lock().unwrap().deref_mut(),
                );
                add_grid(
                    half_height,
                    PlateFormat::W96,
                    &self.main_state.transfers,
                    plate_tool_lib::plate::PlateType::Destination,
                    &ordered_ids,
                    &self.transfer_region_cache,
                    Some(&self.current_transfer_state),
                    ui,
                    self.destination_plate_state.lock().unwrap().deref_mut(),
                );
            });
        });
    }
}
